library tagser;

part 'src/attribute.dart';
part 'src/const.dart';
part 'src/error.dart';
part 'src/messages.dart';
part 'src/result.dart';
part 'src/states.dart';
part 'src/tag.dart';
part 'src/utils.dart';

class Tagser {
  List<TagserState> _stack;
  int _line = 0; // template lining support
  int _symbol = 0;

  List<Tag> parse(String source) {
    _stack = [];
    _stack.add(RootState(null));

    if (source != null && source.isNotEmpty) {
      final lines = source.split('\n');

      for (var l = 0; l < lines.length; l++) {
        _line = l;
        final line = lines[l];

        for (var i = 0; i < line.length; i++) {
          //print('Current state is: ${_stack.last}');
          _symbol = i;
          final charCode = line.codeUnitAt(i);

          //print('Pocessing char "${String.fromCharCode(charCode)}"; State - ${_stack.last.toString()}');
          //print("char: ${line[i]}");

          process(ProcessMessage(charCode: charCode));
        }

        //if (l < lines.length - 1) {
          //process(ProcessMessage(charCode: CHARS.ENTER));
        //}
      }

      process(ProcessMessage(charCode: CHAR_EOS));

      final state = _stack.last;

      if (_stack.length != 1 || !(state is RootState) ) {
        throw Exception(
            'Source document malformed.');
      } else {
        return (state as RootState).tags ;
      }
    }

    return null;
  }

  void process(TagserMessage msg) {
    final state = _stack.last;

    if (state != null && state.canAcceptMessage(msg)) {
      final res = state.processMessage(msg);

      if (res != null) {
        processResult(res);
      }
    }
  }

  void processResult(TagserResult r) {
    if (r.pop == true) {
      pop();
    }

    if (r.state != null) {
      _stack.add(r.state);
    }

    if (r.message != null) {
      process(r.message);
    }

    if (r.err != null) {
      final e = 'Error (${r.err.code}) on $currentLine:$_symbol ${r.err.text}';

      print(e);

      throw Exception(e);
    }
  }

  void pop() {
    _stack.removeLast();
  }

  int get currentLine {
    return _line + 1;
  }
}
